from typing import Dict, Sequence

from eotransform.protocol.transformer import PredicatedTransformer


def transform_all_dict_elems(d: Dict, pred_trans: PredicatedTransformer):
    return {k: _transform_element_in_collection(d, k, pred_trans) for k in d}


def transform_all_sequence_elems(s: Sequence, pred_trans: PredicatedTransformer):
    if isinstance(s, list):
        return [_transform_element_in_collection(s, i, pred_trans) for i, _ in enumerate(s)]
    elif isinstance(s, tuple):
        return tuple(_transform_element_in_collection(s, i, pred_trans) for i, _ in enumerate(s))


def _transform_element_in_collection(collection, index, pred_trans):
    v = collection[index]
    if pred_trans.is_applicable(index):
        return pred_trans.apply(index, v)
    if isinstance(v, dict):
        return transform_all_dict_elems(v, pred_trans)
    if isinstance(v, (list, tuple)):
        return transform_all_sequence_elems(v, pred_trans)
    return v



import logging
import os
import time

import pytest
from pytest import approx

from eotransform.utilities.profiling import PerformanceClock, PERFORMANCE_LOG_ENV_VAR, clock_performance, log


@pytest.fixture
def clock():
    return PerformanceClock()


@pytest.fixture
def set_logging_active():
    try:
        os.environ[PERFORMANCE_LOG_ENV_VAR] = "1"
        yield
    finally:
        del os.environ[PERFORMANCE_LOG_ENV_VAR]


def test_measure_scope(clock, slow_factor):
    with clock.measure():
        time.sleep(0.01 * slow_factor)

    assert clock.mean_measures == approx(0.01 * slow_factor, abs=0.009 * slow_factor)


def test_measure_multiple_scopes_and_provide_mean(clock, slow_factor):
    with clock.measure():
        time.sleep(0.01 * slow_factor)
    with clock.measure():
        time.sleep(0.03 * slow_factor)

    assert clock.mean_measures == approx(0.02 * slow_factor, abs=0.005 * slow_factor)


def test_drop_first(clock, slow_factor):
    with clock.measure():
        time.sleep(0.05 * slow_factor)

    with clock.measure():
        time.sleep(0.01 * slow_factor)

    assert clock.drop_first().mean_measures == approx(0.01 * slow_factor, abs=0.005 * slow_factor)


def test_total_measures(clock, slow_factor):
    for _ in range(5):
        with clock.measure():
            time.sleep(0.01 * slow_factor)

    assert clock.total_measures == approx(0.05 * slow_factor, abs=0.01 * slow_factor)


def test_clock_can_have_a_name():
    assert PerformanceClock("a name").name == "a name"


def test_performance_clock_decorators(caplog, set_logging_active):
    @clock_performance("process")
    def some_process():
        time.sleep(0.01)

    @clock_performance("io")
    def some_io():
        time.sleep(0.03)

    log("process").every(10)
    log("io").every(1)

    with caplog.at_level(logging.INFO):
        for _ in range(10):
            some_process()
            some_io()
        assert sum("some_io" in r.msg for r in caplog.records) == 10
        assert sum("some_process" in r.msg for r in caplog.records) == 1


def test_performance_clock_on_class_methods(caplog, set_logging_active):
    class SomeClass:
        @clock_performance("process")
        def some_method(self):
            time.sleep(0.01)

    log("process").every(1)

    with caplog.at_level(logging.INFO):
        obj = SomeClass()
        obj.some_method()
        assert sum("some_method" in r.msg for r in caplog.records) == 1


def test_ignore_performance_logging_if_environment_variable_is_not_set(caplog):
    @clock_performance("process")
    def some_process():
        time.sleep(0.01)

    log("process").every(1)

    with caplog.at_level(logging.INFO):
        some_process()

    assert len(caplog.records) == 0



import time

from pytest import approx

from eotransform.protocol.transformer import Transformer
from eotransform.transformers.with_performance_clock import WithPerformanceClock
from eotransform.utilities.profiling import PerformanceClock


class ATransformerThatTakes(Transformer):
    def __init__(self, seconds):
        self._seconds = seconds

    def __call__(self, x):
        time.sleep(self._seconds)
        return x


def test_clock_runtime_performance_of_transform(slow_factor):
    clock = PerformanceClock()
    transform = WithPerformanceClock(ATransformerThatTakes(seconds=0.1 * slow_factor), clock)
    transform(0)
    assert clock.total_measures == approx(0.1 * slow_factor, abs=0.01 * slow_factor)



from abc import ABC
from typing import TypeVar

from eotransform.protocol.transformer import Transformer
from eotransform.result import Result

T = TypeVar('T')
E = TypeVar('E')


class ResultTransformation(Transformer[Result[T, E], Result[T, E]], ABC):
    ...


class ApplyToOkResult(ResultTransformation):
    def __init__(self, transformation):
        self._transformation = transformation

    def __call__(self, x: Result[T, E]) -> Result[T, E]:
        if x.is_error():
            return x
        return Result.ok(self._transformation(x.unwrap()))


class Unwrap(ResultTransformation):
    def __call__(self, x: Result[T, E]) -> T:
        return x.unwrap()



from concurrent.futures import ThreadPoolExecutor

import pytest

from doubles import AssertIntStream, SourceStub, ProcessStub, SinkSpy
from eotransform.streamed_process import streamed_process
from eotransform.utilities.profiling import PerformanceClock
from trivial_implementations import Add

EPSILON = 1 / 32


@pytest.fixture
def executor():
    with ThreadPoolExecutor(max_workers=2) as e:
        yield e


def test_process_source_and_put_in_sink(executor):
    out_sink = AssertIntStream(expected=[1, 2, 3])
    streamed_process([0, 1, 2], Add(1), out_sink, executor)
    out_sink.assert_all_visited()


def test_interleave_processing_and_sour_and_sink_operations(executor, slow_factor):
    """
    Testing interleaving INput IO, PRocess and OuTput IO:
    IN->IN->IN
      ->PR->PR->PR
          ->OT->OT->OT
    """
    source = make_io_source(n=3, operation_time=(1 / 16) * slow_factor)
    process = make_process(operation_time=(1 / 16) * slow_factor)
    sink = make_io_sink(operation_time=(1 / 16) * slow_factor)

    streamed_process_clock = PerformanceClock()
    with streamed_process_clock.measure():
        streamed_process(source, process, sink, executor)

    assert streamed_process_clock.mean_measures < (5 / 16 + EPSILON) * slow_factor


def make_io_source(n, operation_time=0.0, raises_error_at=None):
    return SourceStub(n, operation_time, raises_error_at)


def make_process(operation_time=0.0, raises_error_at=None):
    return ProcessStub(operation_time, raises_error_at)


def make_io_sink(operation_time=0.0, raises_error_at=None):
    return SinkSpy(operation_time, raises_error_at)


class SourceError(RuntimeError):
    ...


@pytest.mark.parametrize("err_idx", [0, 1, 2])
def test_hand_through_loading_errors(executor, err_idx):
    error_source = make_io_source(n=3,
                                  raises_error_at={err_idx: SourceError("something bad has happened during loading.")})
    with pytest.raises(SourceError):
        streamed_process(error_source, make_process(), make_io_sink(), executor, loading_timeout=0.01)


class SinkError(RuntimeError):
    ...


@pytest.mark.parametrize("err_idx", [0, 1, 2, 3])
def test_hand_through_storing_errors(executor, err_idx):
    error_sink = make_io_sink(raises_error_at={err_idx: SinkError("something bad has happened during storing.")})
    with pytest.raises(SinkError):
        streamed_process(make_io_source(n=4), make_process(), error_sink, executor, storing_timeout=0.01)


class ProcessError(RuntimeError):
    ...


@pytest.mark.parametrize("err_idx", [0, 1, 2, 3])
def test_cleanup_properly_when_process_errors(executor, err_idx):
    error_process = make_process(
        raises_error_at={err_idx: ProcessError("something bad has happened during processing.")})
    with pytest.raises(ProcessError):
        streamed_process(make_io_source(n=4), error_process, make_io_sink(), executor)


def test_sink_stores_last_successful_values_after_process_error(executor):
    error_process = make_process(raises_error_at={2: ProcessError("something bad happened at the end.")})
    sink = make_io_sink()
    try:
        streamed_process(make_io_source(n=3), error_process, sink, executor)
    except ProcessError:
        ...
    assert sink.received_values == [0, 1]


def test_sink_stores_last_successful_values_after_loading_error(executor):
    error_source = make_io_source(n=3, raises_error_at={2: SourceError("something bad happened at the end.")})
    sink = make_io_sink()
    try:
        streamed_process(error_source, make_process(), sink, executor, loading_timeout=0.01)
    except SourceError:
        ...
    assert sink.received_values == [0, 1]



from typing import TypeVar

from eotransform.protocol.sink import Sink
from eotransform.utilities.profiling import PerformanceClock

ClockedSinkT = TypeVar("ClockedSinkT")


class WithPerformanceClock(Sink[ClockedSinkT]):
    def __init__(self, wrapped: Sink, clock: PerformanceClock):
        self._wrapped = wrapped
        self._clock = clock

    def __call__(self, x: ClockedSinkT) -> None:
        with self._clock.measure():
            return self._wrapped(x)



from typing import Generic, Optional, TypeVar, Set, Type

T = TypeVar('T')
E = TypeVar('E')


class Result(Generic[T, E]):
    """
    Allows to hand exceptions across thread boundaries conveniently by wrapping them into a result object. Use the
    classmethod factories to produce expressive code i.e.:

    >>> def everything_went_fine():
    ...     return Result.ok(42)

    >>> def something_went_wrong():
    ...     return Result.error(RuntimeError("Something bad has happened"))

    >>> everything_went_fine().unwrap()
    42
    >>> something_went_wrong().is_error()
    True
    >>> something_went_wrong().unwrap()
    Traceback (most recent call last):
      ...
    RuntimeError: Something bad has happened
    """

    def __init__(self, value: Optional[T] = None, error: Optional[E] = None, ignored: Optional[E] = None):
        """
        @param value: value of a valid result
        @param error: error of an invalid result
        @param ignored: ignored error
        """
        self._value = value
        self._error = error
        self._ignored = ignored

    @classmethod
    def ok(cls, value: T) -> "Result":
        """
        Factory method creating a valid Result object
        @param value: value of the valid result
        @return: Result object containing the value

        >>> Result.ok(42)
        Result(value=42)
        """
        return cls(value=value)

    @classmethod
    def error(cls, error: E) -> "Result":
        """
        Factory method creating an invalid Result object
        @param error: error which caused the result to be invalid
        @return: Result object containing an error

        >>> Result.error(RuntimeError("Something bad has happened"))
        Result(error=RuntimeError('Something bad has happened'))
        """
        return cls(error=error)

    @classmethod
    def ignored(cls, error: E) -> "Result":
        """
        Factory method creating an error Result object where the error has been ignored
        @param error: error which has been ignored
        @return: Result object containing an ignored error

        >>> Result.ignored(RuntimeError("Something bad has happened"))
        Result(ignored=RuntimeError('Something bad has happened'))
        """
        return cls(ignored=error)

    def unwrap(self) -> T:
        """
        Extract the value of a valid Result object, and throws the set error if it is invalid i.e.:
        @return: value of valid object

        >>> Result.ok(42).unwrap()
        42
        >>> Result.error(RuntimeError("Something bad has happened")).unwrap()
        Traceback (most recent call last):
          ...
        RuntimeError: Something bad has happened
        """
        if self.is_error():
            raise self._error
        return self._value

    def ignore(self, exceptions: Set[Type[E]]) -> "Result":
        """
        Ignore specified exception errors and return new Result object with ignored errors if they match.
        @param exceptions: set of exception types to be ignored
        @return: new result object with exceptions ignored if they match

        >>> Result.error(RuntimeError("An error to be ignored")).ignored({RuntimeError}).unwrap()

        """
        if self.is_error() and type(self._error) in exceptions:
            return Result.ignored(self._error)
        return self

    def is_error(self) -> bool:
        """
        Returns true if an error has been set, and the Result object is invalid
        @return: boolean
        """
        return self._error is not None

    def is_ignored(self) -> bool:
        """
        Returns true if an error has been ignored, and the Result object is invalid
        @return: boolean
        """
        return self._ignored is not None

    def report_error(self) -> str:
        """
        Report the error's string representation.
        @return: string representing the error

        >>> Result.error(RuntimeError("Error report")).report_error()
        'Error report'
        """
        return str(self._error)

    def __repr__(self):
        if self.is_error():
            return f"Result(error={self._error!r})"
        elif self.is_ignored():
            return f"Result(ignored={self._ignored!r})"
        else:
            return f"Result(value={self._value!r})"



from eotransform.transformers.apply_to_enumeration import ApplyToEnumeration
from trivial_implementations import Add, ReplaceWithString


def test_also_works_with_tuples():
    assert tuple(ApplyToEnumeration({
        0: ReplaceWithString("the answer:"),
        1: Add(40)})((1, 2, 3))) == ("the answer:", 42, 3)



import pytest
from eotransform.result import Result
from eotransform.sinks.result import SinkUnwrapped

from tests.helpers.doubles import SinkSpy


def test_sink_unwrapped_result():
    sink = SinkSpy()
    unwrapped_sink = SinkUnwrapped(sink)
    unwrapped_sink(Result.ok(42))
    assert sink.received_values == [42]
    with pytest.raises(AssertionError):
        unwrapped_sink(Result.error(AssertionError("an error occurred")))


def test_sink_unwrapped_result_ignores_specified_exceptions():
    class SomeTestError(AssertionError):
        ...

    sink = SinkSpy()
    unwrapped_sink = SinkUnwrapped(sink, ignore_exceptions={SomeTestError})
    unwrapped_sink(Result.error(SomeTestError("an error to be ignored")))
    assert sink.received_values == []



from typing import Iterable, TypeVar

from eotransform.protocol.transformer import Transformer

ComposedT = TypeVar('ComposedT')
ComposedU = TypeVar('ComposedU')


class Compose(Transformer[ComposedT, ComposedU]):
    """
    Composing a sequence of transformations together

    >>> class PlusOne(Transformer[int, int]):
    ...     def __call__(self, x):
    ...         return x + 1
    ...
    >>> Compose([PlusOne(), PlusOne()])(0)
    2
    """

    def __init__(self, transformations: Iterable[Transformer]):
        """
        :param transformations: Multiple transformations applied to the input data
        """
        self._transformations = transformations

    def __call__(self, x: ComposedT) -> ComposedU:
        for t in self._transformations:
            x = t(x)
        return x



from abc import abstractmethod
from typing import TypeVar, Generic

TransformerT = TypeVar('TransformerT')
TransformerU = TypeVar('TransformerU')


class Transformer(Generic[TransformerT, TransformerU]):
    @abstractmethod
    def __call__(self, x: TransformerT) -> TransformerU:
        ...


PredicatedTransformerKey = TypeVar('PredicatedTransformerKey')
PredicatedTransformerIn = TypeVar('PredicatedTransformerIn')
PredicatedTransformerOut = TypeVar('PredicatedTransformerOut')


class PredicatedTransformer(Generic[PredicatedTransformerKey, PredicatedTransformerIn, PredicatedTransformerOut]):
    @abstractmethod
    def is_applicable(self, k: PredicatedTransformerKey) -> bool:
        ...

    @abstractmethod
    def apply(self, k: PredicatedTransformerKey, x: PredicatedTransformerIn) -> PredicatedTransformerOut:
        ...



import logging
import os
import time
from contextlib import contextmanager
from functools import wraps
from typing import Optional

PERFORMANCE_LOG_ENV_VAR = 'CORRELATION_MAP_PERFORMANCE_LOG'
NANOSECONDS_PER_SECOND = 1e9

class PerformanceClock:
    def __init__(self, name: Optional[str] = ""):
        self._measures = []
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @contextmanager
    def measure(self):
        start = time.perf_counter_ns()
        yield
        self._measures.append((time.perf_counter_ns() - start) / NANOSECONDS_PER_SECOND)

    @property
    def total_measures(self):
        return sum(self._measures)

    @property
    def mean_measures(self):
        return self.total_measures / len(self._measures)

    def drop_first(self):
        self._measures = self._measures[1:]
        return self


_LOGGING_FREQUENCIES = dict()


def log(domain: str):
    class _FrequencySetter:
        def __init__(self, domain):
            self._domain = domain

        def every(self, frequency: int):
            _LOGGING_FREQUENCIES[self._domain] = frequency

    return _FrequencySetter(domain)


def clock_performance(domain: str):
    def _clock_performance(func):
        if not os.environ.get(PERFORMANCE_LOG_ENV_VAR, False):
            return func

        clock = PerformanceClock()
        log_n = 0
        freq = None

        @wraps(func)
        def _decorator(*args, **kwargs):
            nonlocal freq
            nonlocal log_n
            nonlocal clock
            if freq is None:
                freq = _LOGGING_FREQUENCIES.get(domain, 0)
            if freq == 0:
                return func(*args, **kwargs)

            with clock.measure():
                response = func(*args, **kwargs)

            log_n += 1
            if log_n % freq == 0:
                logging.info(f"{domain}::{func.__name__}: {clock.mean_measures}")
            return response

        return _decorator

    return _clock_performance



from abc import abstractmethod
from typing import TypeVar, Generic

SinkT = TypeVar('SinkT')


class Sink(Generic[SinkT]):
    """
    Basic protocol to consume data
    """

    @abstractmethod
    def __call__(self, x: SinkT) -> None:
        ...



from typing import Callable

from eotransform.protocol.sink import Sink, SinkT


class SinkFiltered(Sink[SinkT]):
    """
    Allows to filter items for passing them to the wrapped sink.
    >>> class RecordingSink(Sink[int]):
    ...     def __init__(self):
    ...          self.record = []
    ...     def __call__(self, x: int) -> None:
    ...          self.record.append(x)
    >>> wrapped_sink = RecordingSink()
    >>> filtered_sink = SinkFiltered(wrapped_sink, lambda x: x % 2 == 1)
    >>> for i in range(4):
    ...     filtered_sink(i)
    >>> wrapped_sink.record
    [1, 3]
    """
    def __init__(self, wrapped_sink: Sink[SinkT], predicate: Callable[[SinkT], bool]):
        """
        :param wrapped_sink: the sink the filtered items are passed to
        :param predicate: callable which determines if an item is passed to the wrapped sink
        """
        self._wrapped_sink = wrapped_sink
        self._predicate = predicate

    def __call__(self, x: SinkT) -> None:
        if self._predicate(x):
            self._wrapped_sink(x)



from typing import Iterable, Tuple, TypeVar

from eotransform.protocol.transformer import Transformer

TupleElemT = TypeVar('TupleElemT')


class ToTuple(Transformer[Iterable[TupleElemT], Tuple[TupleElemT, ...]]):
    """
    Fill element of an iterable stream into a tuple
    >>> ToTuple()([0, 1, 2])
    (0, 1, 2)
    """

    def __call__(self, x: Iterable[TupleElemT]) -> Tuple[TupleElemT, ...]:
        return tuple(x)



from itertools import chain
from typing import TypeVar, Iterable, Iterator

from eotransform.protocol.transformer import Transformer

ChainT = TypeVar('ChainT')


class Chain(Transformer[Iterable[ChainT], Iterator[ChainT]]):
    """
    Chain multiple iterables together into one iterator, i.e.:

    >>> list(Chain()([[0, 1], [2]]))
    [0, 1, 2]
    """

    def __call__(self, x: Iterable[ChainT]) -> Iterator[ChainT]:
        return chain.from_iterable(x)



from abc import abstractmethod
from typing import Any, Optional

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol

from eotransform.protocol.sink import Sink


class Reporter(Protocol):
    @abstractmethod
    def update(self, n: Optional[int] = 1) -> None:
        ...


class SinkToProgressReport(Sink[Any]):
    def __init__(self, reporter: Reporter):
        self._reporter = reporter

    def __call__(self, x: Any) -> None:
        self._reporter.update()



from itertools import repeat
from typing import TypeVar, Iterator

from eotransform.protocol.transformer import Transformer

RepeatedT = TypeVar('RepeatedT')


class Repeat(Transformer[RepeatedT, Iterator[RepeatedT]]):
    """
    Repeat the input n times and return an iterator which can for instance be used to create a collection, such as list.
    >>> list(Repeat(3)(42))
    [42, 42, 42]
    """

    def __init__(self, n: int):
        """
        :param n: Number of times the input is repeated.
        """
        self._n = n

    def __call__(self, x: RepeatedT) -> Iterator[RepeatedT]:
        return repeat(x, self._n)



from eotransform.collection_transformation import transform_all_dict_elems
from eotransform.protocol.transformer import PredicatedTransformer, PredicatedTransformerKey, PredicatedTransformerIn, \
    PredicatedTransformerOut


class PredicatedTransformerStub(PredicatedTransformer):

    def is_applicable(self, k: PredicatedTransformerKey) -> bool:
        return k == "transformable"

    def apply(self, k: PredicatedTransformerKey, x: PredicatedTransformerIn) -> PredicatedTransformerOut:
        return 42


def test_transform_nested_collection():
    collection = dict(nested=dict(transformable="What is the answer?", list=["stable", "stable"]),
                      list=["stable", "stable"], tuple=("stable", dict(transformable="What is the answer?")),
                      transformable="The value is not important")
    assert transform_all_dict_elems(collection, PredicatedTransformerStub()) == \
           dict(nested=dict(transformable=42, list=["stable", "stable"]),
                list=["stable", "stable"], tuple=("stable", dict(transformable=42)),
                transformable=42)



from eotransform.sinks.filtered import SinkFiltered

from tests.helpers.doubles import SinkSpy


def test_sink_only_items_where_lambda_returns_true():
    def is_even(x):
        return x % 2 == 0

    sink_wrapped = SinkSpy()
    sink_filtered = SinkFiltered(sink_wrapped, is_even)
    [sink_filtered(i) for i in range(5)]
    assert sink_wrapped.received_values == [0, 2, 4]



from typing import TypeVar

from eotransform.protocol.sink import Sink, SinkT
from eotransform.result import Result

ErrT = TypeVar("ErrT")


class SinkUnwrapped(Sink[Result[SinkT, ErrT]]):
    """
    Unwrap a result object before passing it onto the wrapped sink. Also allows for automatically ignoring exceptions.
    >>> class RecordingSink(Sink[int]):
    ...     def __init__(self):
    ...          self.record = []
    ...     def __call__(self, x: int) -> None:
    ...          self.record.append(x)
    >>> sink = RecordingSink()
    >>> SinkUnwrapped(sink)(Result.ok(42))
    >>> sink.record
    [42]
    >>> SinkUnwrapped(sink, ignore_exceptions={RuntimeError})(Result.error(RuntimeError("An error to be ignored")))
    >>> sink.record
    [42]
    """
    def __init__(self, wrapped_sink: Sink[SinkT], ignore_exceptions=None):
        """
        @param wrapped_sink: sink to which the unwrapped object is passed
        @param ignore_exceptions: set of exception types to be ignored
        """
        self._wrapped_sink = wrapped_sink
        self._ignore_exceptions = ignore_exceptions or set()

    def __call__(self, x: Result[SinkT, ErrT]) -> None:
        x = x.ignore(self._ignore_exceptions)
        if not x.is_ignored():
            self._wrapped_sink(x.unwrap())



from typing import TypeVar

from eotransform.protocol.transformer import Transformer
from eotransform.utilities.profiling import PerformanceClock

ClockedT = TypeVar("ClockedT")
ClockedU = TypeVar("ClockedU")


class WithPerformanceClock(Transformer[ClockedT, ClockedU]):
    def __init__(self, wrapped: Transformer, clock: PerformanceClock):
        self._wrapped = wrapped
        self._clock = clock

    def __call__(self, x):
        with self._clock.measure():
            return self._wrapped(x)



from concurrent.futures import ThreadPoolExecutor
from queue import Queue, Empty, Full
from threading import Event
from typing import TypeVar, Iterable, Iterator

from eotransform.protocol.sink import Sink
from eotransform.protocol.transformer import Transformer

StreamedInT = TypeVar('StreamedInT')
StreamedOutT = TypeVar('StreamedOutT')
ReturnPipeT = TypeVar('ReturnPipeT')
InSource = Iterable[StreamedInT]
ProcessTransformer = Transformer[StreamedInT, StreamedOutT]
OutSink = Sink[StreamedOutT]


class ReturnPipeIterator(Iterator[ReturnPipeT]):
    def __init__(self, pipe: Queue, timeout: float):
        self._pipe = pipe
        self._timeout = timeout

    def __next__(self) -> ReturnPipeT:
        while True:
            elem = self._pipe.get(timeout=self._timeout)
            self._pipe.task_done()
            if elem is StopIteration:
                raise elem
            return elem


class ReturnPipeStream(Iterable[ReturnPipeT]):
    def __init__(self, pipe: Queue, timeout: float):
        self._pipe = pipe
        self._timeout = timeout

    def __iter__(self) -> Iterator[ReturnPipeT]:
        return ReturnPipeIterator(self._pipe, self._timeout)


def _producer(source: InSource, pipe: Queue, cancel_event: Event) -> None:
    for elem in source:
        if cancel_event.is_set():
            return
        else:
            pipe.put(elem)

    if not cancel_event.is_set():
        pipe.put(StopIteration)


def _consumer(sink: OutSink, pipe: Queue) -> None:
    for elem in ReturnPipeStream(pipe, 9999):
        sink(elem)


def streamed_process(source: InSource, process: ProcessTransformer, dst_sink: OutSink,
                     executor: ThreadPoolExecutor, loading_timeout: float = 120, storing_timeout: float = 120) -> None:
    """
    Interleaves the data loading from source with the data processing, and finally putting it into the sink. This can be
    used to hide IO processed i.e.:
    thread_0: IN->IN->IN
    thread_1:   ->PR->PR->PR
    thread_2:       ->SK->SK->SK

    :param source: Input data fed to the process transformation
    :param process: Transformation changing the input data to the output data
    :param dst_sink: Sink receiving the output data
    :param executor: Thread pool executor used to spawn async threads for IO hiding
    :param loading_timeout: Timeout for the loading process in seconds (default: 120)
    :param storing_timeout: Timout for the storing process in seconds (default: 120)
    """

    source_pipe = Queue(maxsize=1)
    sink_pipe = Queue(maxsize=1)

    cancel_production_event = Event()
    producer = executor.submit(_producer, source, source_pipe, cancel_production_event)
    consumer = executor.submit(_consumer, dst_sink, sink_pipe)

    try:
        for x in ReturnPipeStream(source_pipe, loading_timeout):
            sink_pipe.put(process(x), timeout=storing_timeout)
    except Empty:
        ...
    finally:
        cancel_production_event.set()
        try:
            sink_pipe.put(StopIteration, timeout=storing_timeout)
        except Full:
            ...
        finally:
            for _ in range(source_pipe.unfinished_tasks):
                source_pipe.get_nowait()
                source_pipe.task_done()

            producer.result(timeout=loading_timeout)
            consumer.result(timeout=storing_timeout)
            _unsafe_pipe_clear(source_pipe)
            source_pipe.join()
            sink_pipe.join()


def _unsafe_pipe_clear(pipe: Queue):
    pipe.queue.clear()
    for _ in range(pipe.unfinished_tasks):
        pipe.task_done()



from typing import Dict, Iterable, Iterator, TypeVar

from eotransform.protocol.transformer import Transformer

InSelectT = TypeVar('InSelectT')
OutSelectT = TypeVar('OutSelectT')


class ApplyToEnumeration(Transformer[Iterable[InSelectT], Iterator[OutSelectT]]):
    """
    Apply a transformations to the element of an enumeration selected by the index
    >>> class Plus(Transformer[int, int]):
    ...     def __init__(self, value):
    ...         self.value = value
    ...
    ...     def __call__(self, x):
    ...         return x + self.value
    ...
    >>> list(ApplyToEnumeration({1: Plus(1), 2: Plus(2)})([0, 1, 2]))
    [0, 2, 4]
    """

    def __init__(self, transforms: Dict[int, Transformer[Iterable[InSelectT], Iterator[OutSelectT]]]):
        """
        :param transforms: Dictionary of transformations applied to the index specified as key
        """
        self._transforms = transforms

    def __call__(self, x: Iterable[InSelectT]) -> Iterator[OutSelectT]:
        for i, elem in enumerate(x):
            if i in self._transforms:
                yield self._transforms[i](elem)
            else:
                yield elem



from typing import Iterator, Sequence, TypeVar

from eotransform.protocol.transformer import Transformer

MapInT = TypeVar('MapInT')
MapOutT = TypeVar('MapOutT')


class Map(Transformer[Sequence[MapInT], Iterator[MapOutT]]):
    """
    Applying a transformation to a sequence of input data, i.e.:

    >>> class PlusOne(Transformer[int, int]):
    ...     def __call__(self, x):
    ...         return x + 1
    ...
    >>> list(Map(PlusOne())([0, 1, 2]))
    [1, 2, 3]
    """

    def __init__(self, map_transformer: Transformer[MapInT, MapOutT]):
        """
        :param map_transformer: Transformer applied to each element in the sequence
        """
        self._map_transformer = map_transformer

    def __call__(self, x: Sequence[MapInT]) -> Iterator[MapOutT]:
        return map(self._map_transformer, x)



import pytest

from eotransform.result import Result
from eotransform.transformers.result import ApplyToOkResult, Unwrap


def test_apply_to_ok_result_transforms_ok_result():
    x = Result.ok(42)
    assert ApplyToOkResult(lambda r: r + 1)(x).unwrap() == 43


def test_apply_to_ok_result_hands_through_error_without_applying_any_transformation():
    def a_transform(_):
        assert False

    x = Result.error(AssertionError("an error occurred"))
    x = ApplyToOkResult(a_transform)(x)
    with pytest.raises(AssertionError):
        x.unwrap()


def test_unwrap_result():
    assert Unwrap()(Result.ok(42)) == 42
    with pytest.raises(AssertionError):
        Unwrap()(Result.error(AssertionError("an error occured")))



from typing import TypeVar

from eotransform.protocol.transformer import Transformer

IdentityT = TypeVar('IdentityT')


class Identity(Transformer[IdentityT, IdentityT]):
    """
    The identity transformation
    >>> Identity()(42)
    42
    """

    def __call__(self, x: IdentityT) -> IdentityT:
        return x



import pytest

from eotransform.sinks.sink_to_progress_report import Reporter, SinkToProgressReport


class ReporterSpy(Reporter):
    def __init__(self):
        self.num_update_calls_received = 0

    def update(self, n=1) -> None:
        self.num_update_calls_received += 1


@pytest.fixture
def reporter():
    return ReporterSpy()


def test_report_iteration(reporter):
    sink = SinkToProgressReport(reporter)
    sink(42)
    assert reporter.num_update_calls_received == 1
    sink(42)
    assert reporter.num_update_calls_received == 2



from typing import TypeVar, Callable

from eotransform.protocol.transformer import Transformer, TransformerT, TransformerU

ApplyT = TypeVar('ApplyT')
ApplyU = TypeVar('ApplyU')


class Apply(Transformer[ApplyT, ApplyU]):
    """
    Applies an arbitrary function to the input which takes the input as argument and returns a transformed output

    >>> Apply(lambda x: f"The answer is: {x}")(42)
    'The answer is: 42'
    """

    def __init__(self, fn: Callable[[ApplyT], ApplyU]):
        self._fn = fn

    def __call__(self, x: TransformerT) -> TransformerU:
        return self._fn(x)
